"""The Mimeo Resources package.

This package contains all resources used in Mimeo, such as countries
and cities data or logging configuration file. It also includes
custom exceptions module:
* exc
    The Mimeo Resources Exceptions module.

To use this package, simply import the desired exception:
    from mimeo.resources.exc import ResourceNotFoundError
"""
